package com.kfi.dm.gui.core.event;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import com.google.common.collect.MapMaker;
import com.kfi.dm.gui.core.event.exception.DuplicateSubscriberMethodException;
import com.kfi.dm.gui.core.event.exception.DuplicateSubscriptionException;
import com.kfi.dm.gui.core.event.exception.EventClassNotAnnotatedException;
import com.kfi.dm.gui.core.event.exception.EventNotSubscribedException;
import com.kfi.dm.gui.core.event.exception.IncompatibleSubscriberMethodException;
import com.kfi.dm.gui.core.event.exception.NoSubscriberMethodFoundException;

public class EventBus {

	private final Map<Class<?>, Set<Object>> eventSubscriberMap = new MapMaker().weakKeys().weakValues().makeMap();
	private final Map<Object, Map<Class<?>, Method>> subscriberEventMethodMap = new MapMaker().weakKeys().weakValues().makeMap();

	public void addSubscriber(final Object subscriber) {
		assertNotNull(subscriber);
		Class<?> subscriberClass = subscriber.getClass();
		Log.log("Adding subscriber class " + subscriberClass);
		
		if (subscriberEventMethodMap.containsKey(subscriber)) {
			throw new DuplicateSubscriptionException(subscriberClass);
		}
		
		Map<Class<?>, Method> eventMethodMap = scanForSubscriberMethods(subscriberClass);
		if (eventMethodMap.size() <= 0) {
			throw new NoSubscriberMethodFoundException(subscriberClass);
		}
		
		for (Map.Entry<Class<?>, Method> entry : eventMethodMap.entrySet()) {
			Class<?> eventClass = entry.getKey();
			
			Set<Object> subscriptionSet = eventSubscriberMap.get(eventClass);
			if (subscriptionSet == null) {
				subscriptionSet = new HashSet<Object>();
				eventSubscriberMap.put(eventClass, subscriptionSet);
			}
			subscriptionSet.add(subscriber);
		}
		
		subscriberEventMethodMap.put(subscriber, eventMethodMap);
	}

	private Map<Class<?>, Method> scanForSubscriberMethods(Class<?> subscriberClass) {
		Map<Class<?>, Method> subscriberMap = new HashMap<Class<?>, Method>();
		for (Method method : subscriberClass.getMethods()) {
			final EventSubscriber annotation = method.getAnnotation(EventSubscriber.class);

			if (annotation != null) {
				Class<?> eventClass = annotation.value();
				if (eventClass.getAnnotation(Event.class) == null) {
					throw new EventClassNotAnnotatedException(eventClass);
				}

				final Class<?>[] parameterTypes = method.getParameterTypes();
				if (parameterTypes.length != 1 || !parameterTypes[0].equals(eventClass)) {
					throw new IncompatibleSubscriberMethodException(subscriberClass, method, eventClass);
				}
				
				if (subscriberMap.containsKey(eventClass)) {
					throw new DuplicateSubscriberMethodException(subscriberClass, method, subscriberMap.get(eventClass));
				} else {
					subscriberMap.put(annotation.value(), method);
					Log.log("    Subscriber method found: " + method.getName() + "for event " + eventClass.getName());
				}
			}
		}
		return subscriberMap;
	}

	public void removeSubscriber(final Object subscriber) {
		assertNotNull(subscriber);
		Class<?> subscriberClass = subscriber.getClass();
		Log.log("Removing subscriber class " + subscriberClass.getName());

		Set<Class<?>> eventSet = subscriberEventMethodMap.get(subscriber).keySet();
		for (Class<?> eventClass : eventSet) {
			eventSubscriberMap.get(eventClass).remove(subscriber);
		}
		
		subscriberEventMethodMap.remove(subscriber);
	}

	public void fire(final Object event) {
		assertNotNull(event);
		Log.log("Firing " + event + "...");

		Class<?> eventClass = event.getClass();
		Set<Object> subscriberSet = eventSubscriberMap.get(event.getClass());

		if (subscriberSet == null || subscriberSet.size() <= 0) {
			throw new EventNotSubscribedException(event.getClass());
		}

		for (final Object subscriber : subscriberSet) {
			try {
				if (subscriber == null) {
					removeSubscriber(subscriber);
				} else {
					Log.log("    triggering subscriber " + subscriber.getClass());
					subscriberEventMethodMap.get(subscriber).get(eventClass).invoke(subscriber, event);
				}
			} catch (final IllegalArgumentException e) {
				e.printStackTrace();
			} catch (final IllegalAccessException e) {
				e.printStackTrace();
			} catch (final InvocationTargetException e) {
				e.printStackTrace();
			}
		}
	}

	private void assertNotNull(final Object... args) {
		for (int i = 0; i < args.length; i++) {
			if (args[i] == null)
				throw new NullPointerException("Argument with index " + i + " was null.");
		}
	}

	public void enableLogging() {
		Log.setLogging(true);
	}

	public void disableLogging() {
		Log.setLogging(false);
	}
}
